# Add tests here 

import unittest
from elara import LRU

class RunTests(unittest.TestCase):
    def setUp(self):
        self.lru = LRU()
        self.lru.push('key1')
        self.lru.push('key2')
        
    def test_push(self):
        self.assertEqual(self.lru.all(), ['key2', 'key1'])
        
    def test_peek(self):
        self.assertEqual(self.lru.peek(), 'key2')
        
    def test_touch(self):
        self.lru.touch('key1')
        self.assertEqual(self.lru.peek(), 'key1')
        
    def test_clear(self):
        self.lru.clear()
        self.assertEqual(self.lru.all(), [])
        
    def test_rem(self):
        self.lru.rem('key2')
        self.assertEqual(self.lru.all(), ['key1'])
    
    def test_pop(self):
        self.assertEqual(self.lru.pop(), 'key1')
        self.assertEqual(len(self.lru.all()), 1)


'''
Copyright (c) 2021, Saurabh Pujari
All rights reserved.

This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree.
'''

# Add all custom exception classes here 

class FileAccessError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'Error -> {self.message}'

class FileKeyError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'Error -> {self.message}'


'''
Copyright (c) 2021, Saurabh Pujari
All rights reserved.

This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree.
'''

# shared operations
from cryptography.fernet import Fernet
from .elarautil import Util
import json
import os
from .exceptions import FileAccessError, FileKeyError

def retdb(self):
    if self.key:
        return Util.readAndDecrypt(self)
    else:
        return Util.readJSON(self)

def retmem(self):
    return self.db

def retkey(self):
    return self.key

def commit(self):
    self._dump()
    
def exportdb(self, export_path, sort=True):
    db = self.retdb()
    new_export_path = os.path.expanduser(export_path)
    try:
        json.dump(db, open(new_export_path, 'wt'), indent=4, sort_keys=sort)
    except Exception:
        raise FileAccessError("Store JSON error. File path might be wrong")
        

def exportmem(self, export_path, sort=True):
    db = self.retmem()
    new_export_path = os.path.expanduser(export_path)
    try:
        json.dump(db, open(new_export_path, 'wt'), indent=4, sort_keys=sort)
    except Exception:
        raise FileAccessError("Store JSON error. File path might be wrong")


def exportkeys(self, export_path, keys = [], sort=True):
    db = {}
    for key in keys:
        if isinstance(key, str) and self.exists(key):
            db[key] = self.get(key)

    new_export_path = os.path.expanduser(export_path)
    try:
        json.dump(db, open(new_export_path, 'wt'), indent=4, sort_keys=sort)
    except Exception:
        raise FileAccessError("Store JSON error. File path might be wrong")


def securedb(self, key_path=None):
    if self.key:
        self.updatekey(self, key_path)
    else:
        new_key_path = os.path.expanduser(key_path)
        self.db = self.retdb() # Store the current contents

        if os.path.exists(new_key_path):
            if os.stat(new_key_path).st_size == 0: 
                Util.keygen(new_key_path)
            else:
                # os.remove(new_key_path)
                f = open(new_key_path, 'r+')
                f.truncate(0)
                f.close()
                Util.keygen(new_key_path)
        else:
            Util.keygen(new_key_path)

        self.key = Util.readkey(new_key_path)
        Util.encryptAndStore(self)
        return True


# Incomplete function
def updatekey(self, key_path=None):
    if self.key:
        new_key_path = os.path.expanduser(key_path)
        self.db = self.retdb()
        if os.path.exists(new_key_path):
            if os.stat(new_key_path).st_size == 0: 
                Util.keygen(new_key_path)
            else:
                # os.remove(new_key_path)
                f = open(new_key_path, 'r+')
                f.truncate(0)
                f.close()
                Util.keygen(new_key_path)

        else:
            Util.keygen(new_key_path)

        #clear db file and encrypt contents with new key
        f = open(self.path, 'r+')
        f.truncate(0)
        f.close 
        self.key = Util.readkey(new_key_path)
        Util.encryptAndStore(self)
        
    else:
        raise FileKeyError("Update key Failed")

    # write function to check if db file exists/ there is data present in it


# Add tests here 

import unittest
import elara

class RunTests(unittest.TestCase):
    def setUp(self):
        self.db = elara.exe('test.db', False)

    def test_setnx(self):
        self.db.set('key1', 'value1')
        self.assertEqual(self.db.setnx('key1', 'value'), False)
        self.assertEqual(self.db.setnx('key', 'value'), True)
        self.assertEqual(self.db.setnx('key', 'value'), False)
    
    def test_append(self):
        self.db.set('key', 'value')
        self.assertEqual(self.db.append('key', 'data'), 'valuedata')
        self.assertEqual(self.db.append('key', 1), False)

    def test_getset(self):
        self.db.set('key', 'value')
        self.assertEqual(self.db.getset('key', 'newvalue'), 'value')
        self.assertEqual(self.db.getset('key', 1), 'newvalue')
        self.assertEqual(self.db.get('key'), 1)

    def test_mget(self):
        self.db.set('key', 'value')
        self.db.set('key1', 3)
        keys = ['key', 'key1', 'key2']
        self.assertEqual(self.db.mget(keys), ['value', 3])
        self.db.set('key2', 'value2')
        self.assertEqual(self.db.mget(keys), ['value', 3, 'value2'])
        self.assertNotEqual(self.db.mget(keys), ['value', 'value2', 3])

    def test_mset(self):
        new_dict={
            "key1" : [1,2,3],
            1 : "value",
            'key2' : 'value2'
        }
        self.db.mset(new_dict)
        self.assertEqual(self.db.mget(['key1', 1, 'key2']), [[1,2,3], 'value2'])

    def test_msetnx(self):
        self.db.set('key', 'value')
        new_dict={
            "key" : [1,2,3],
            "key2" : "value2",
            'key3' : 'value3'
        }
        self.db.msetnx(new_dict)
        self.assertEqual(self.db.mget(['key', 'key3', 'key2']), ['value', 'value3', 'value2'])



# Add tests here 

import unittest
import elara

class RunTests(unittest.TestCase):
    def setUp(self):
        self.db = elara.exe('test.db', False)

    def test_exe(self):
        res = elara.exe('test.db', False)
        assert res is not None

    def test_get(self):
        self.db.db['key'] = "test"
        res = self.db.db['key']
        self.assertEqual(res, 'test')
    
    def test_set(self):
        self.db.set('key', 'test')
        self.assertEqual(self.db.get('key'), 'test')

    def test_rem(self):
        self.db.set('key', 'test')
        assert 'key' in self.db.db
        self.db.rem('key')
        assert 'key' not in self.db.db

    def test_clear(self):
        self.db.set('key', 'test')
        self.assertEqual(self.db.clear(), True)
        self.assertEqual(self.db.db, {})
        
    def test_cull(self):
        self.db.set('one', 1)
        self.db.set('two', 2)
        self.db.set('three', 3)
        self.db.set('four', 4)
        self.db.cull(25)
        self.assertEqual(len(self.db.retmem()), 3)
        self.db.cull(100)
        self.assertEqual(len(self.db.retmem()), 0)
        self.assertEqual(self.db.cull(-1), False)
        self.assertEqual(self.db.cull(101), False)
        
    def test_incr(self):
        self.db.set('one', 1)
        self.db.incr('one')
        self.assertEqual(self.db.get('one'), 2)
        self.db.incr('one', 3.6)
        self.assertEqual(self.db.get('one'), 5.6)
        self.db.incr('one', 0.0003)
        self.assertEqual(self.db.get('one'), 5.600)
        self.db.incr('one', -1)
        self.assertEqual(self.db.get('one'), 4.600)
        
    def test_decr(self):
        self.db.set('one', 1.35)
        self.db.decr('one')
        self.assertEqual(self.db.get('one'), 0.35)
        self.db.decr('one')
        self.assertEqual(self.db.get('one'), -0.65)


'''
Copyright (c) 2021, Saurabh Pujari
All rights reserved.

This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree.
'''
import os
from .elarautil import Util
from .lru import LRU

class Elara():

    from .strings import (setnx, append, getset, mget, mset, msetnx, slen)
    from .lists import (lnew, lpush, lextend, lindex, lrange, lrem, lpop, llen, lappend, lexists, linsert)
    from .hashtables import (hnew, hadd, haddt, hget, hpop, hkeys, hvals, hexists, hmerge)
    from .shared import (retmem, retdb, retkey, commit, exportdb, exportkeys, exportmem, securedb, updatekey)

    def __init__(self, path, commitdb, key_path = None):
        self.path = os.path.expanduser(path)
        self.commitdb = commitdb 
        self.lru = LRU()

        # Since key file is generated first, invalid token error for pre existing open dbs

        # Load the database key
        if not key_path==None:
            new_key_path = os.path.expanduser(key_path)
            if os.path.exists(new_key_path):
                file = open(new_key_path, 'rb')
                self.key = file.read()
                file.close()
            else:
                self.key = None
        else:
            self.key = None

        # Load the data
        if os.path.exists(path):
            self._load()
        else:
            self.db = {}

    def _load(self):
        if self.key:
            self.db = Util.readAndDecrypt(self)
        else:
            self.db = Util.readJSON(self)

    def _dump(self):
        if self.key:
            Util.encryptAndStore(self) # Enclose in try-catch
        else:
            Util.storeJSON(self)

    def _autocommit(self):
        if self.commitdb:
            self._dump()
    
    def set(self, key, value):
        if isinstance(key, str):
            self.db[key] = value
            self.lru.push(key)
            self._autocommit()
            return True
        else:
            raise Exception
    
    def get(self, key):
        try:
            self.lru.touch(key)
            return self.db[key]
        except KeyError:
            return None
    
    def rem(self, key):
        self.lru.rem(key)
        del self.db[key]
        self._autocommit()
        return True
    
    def clear(self):
        self.lru.clear()
        self.db = {}
        self._autocommit()
        return True
    
    def exists(self, key):
        self.lru.touch(key)
        return key in self.db
    
    def cull(self, percentage=10):
        if 0 <=  percentage <= 100 :
            count = int((percentage/100)*(self.numkeys()))
            print("final count", count)
            for i in range(0, count):
                key = self.lru.pop()
                del self.db[key] 
            self._autocommit()   
            return True 
        else:
            return False   
    
    def getkeys(self):
        return self.lru.all()
    
    def numkeys(self):
        return self.lru.size
    
    def incr(self, key, val=1):
        if self.exists(key):
            data = self.get(key)
            if isinstance(data, (int, float)):
                data += val
                data = round(data, 3)
                self.set(key, data)
            else:
                return False
        else:
            return False
        
    def decr(self, key, val=1):
        if self.exists(key):
            data = self.get(key)
            if isinstance(data, (int, float)):
                data -= val
                data = round(data, 3)
                self.set(key, data)
            else:
                return False
        else:
            return False


'''
Copyright (c) 2021, Saurabh Pujari
All rights reserved.

This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree.
'''

from cryptography.fernet import Fernet
import json
import base64
import os
from .exceptions import FileAccessError, FileKeyError

class Util:
    @staticmethod
    def readJSON(obj):
        try:
            curr_db = json.load(open(obj.path, 'rb'))
        except Exception:
            # print("Read JSON error. File might be encrypted. Run in secure mode.")
            raise FileAccessError("Read JSON error. File might be encrypted. Run in secure mode with key path.")
        return curr_db

    @staticmethod
    def storeJSON(obj):
        try:
            json.dump(obj.db, open(obj.path, 'wt'), indent=4)
        except Exception:
            raise FileAccessError("Store JSON error. File might be encrypted. Run in secure mode with key path.")
    
    @staticmethod
    def readAndDecrypt(obj):
        if obj.key:
            fernet = Fernet(obj.key)
            encrypted_data=""
            try:
                with open(obj.path, 'rb') as file:
                    encrypted_data = file.read()
            except Exception:
                raise FileAccessError("File open & read error")
            decrypted_data = fernet.decrypt(encrypted_data)
            data_bytes = base64.b64decode(decrypted_data)
            data_ascii = data_bytes.decode('ascii')
            data_ascii = data_ascii.replace("'", "\"")
            curr_db = json.loads(data_ascii)
            return curr_db
        else:
            return None

    @staticmethod
    def encryptAndStore(obj):
        if obj.key:
            fernet = Fernet(obj.key)
            db_curr = str(obj.db)
            db_ascii = db_curr.encode('ascii')
            db_byte = base64.b64encode(db_ascii)
            encrypted_data = fernet.encrypt(db_byte)
            try:
                with open(obj.path, 'wb') as file:
                    file.write(encrypted_data)
                    return True
            except Exception:
                raise FileAccessError("File open & write error")
        else:
            return False

    @staticmethod
    def keygen(path):
        try:
            key = Fernet.generate_key()
        except Exception:
            raise FileKeyError("Key generation error")
        try:
            with open(path, 'wb') as file:
                file.write(key)
            file.close()
            return True
        except Exception:
            raise FileAccessError("File open & write error")


    @staticmethod
    def loadkey(path):
        key_path = os.path.expanduser(path)
        # print(key_path)

        if os.path.exists(key_path):
            if os.stat(key_path).st_size == 0: 
                Util.keygen(key_path)
                # else: key exists in file; use that
        else:
            # create file and store keygen
            Util.keygen(key_path)

    @staticmethod
    def readkey(path):
        key_path = os.path.expanduser(path)
        if os.path.exists(key_path):
            file = open(key_path, 'rb')
            key = file.read()
            file.close()
            return key
        else:
            key = None
            return key



# Add tests here 

import unittest
import elara

class RunTests(unittest.TestCase):
    def setUp(self):
        self.db = elara.exe('test.db', False)
        self.db.lnew('newlist')

    def test_lpush(self):
        self.assertEqual(len(self.db.get('newlist')), 0)
        self.db.lpush('newlist', 3)
        self.assertEqual(len(self.db.get('newlist')), 1)
        self.db.lpush('newlist', 4)
        self.assertEqual(len(self.db.get('newlist')), 2)
        new_list = self.db.get('newlist')
        self.assertEqual(new_list[1], 4)
        self.db.lpush('newlist', 'string')
        new_list = self.db.get('newlist')
        self.assertEqual(new_list[2], 'string')
        
    def test_lextend(self):
        self.db.lpush('newlist', 3)
        self.assertEqual(len(self.db.get('newlist')), 1)
        new_list = [4, 5]
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(len(self.db.get('newlist')), 3)
        self.assertEqual(self.db.lextend('newlist2', new_list), False)
        
    def test_lindex(self):
        self.db.lpush('newlist', 3)
        new_list = [4, 5]
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(self.db.lindex('newlist', 0), 3)
        self.assertEqual(self.db.lindex('newlist', 5), False)
        
    def test_lrange(self):
        new_list = [4, 5, 6, 7, 8, 9]
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(self.db.lrange('newlist', 2, 4), [6, 7])
        self.assertEqual(self.db.lrange('newlist', 10, 20), [])
        
    def test_lrem(self):
        new_list = ['cat', 'dog', 'bird']
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(self.db.lrem('newlist', 'dog'), True)
        self.assertEqual(self.db.lrem('newlist', 20), False)
        
    def test_lpop(self):
        new_list = [4, 5, 6, 7, 8, 9]
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(self.db.lpop('newlist', -2), False)
        self.assertEqual(self.db.lpop('newlist'), 9)
        self.assertEqual(len(self.db.get('newlist')), len(new_list) -1)
        self.assertEqual(self.db.lpop('list'), False)
        
    def test_llen(self):
        new_list = [4, 5, 6, 7, 8, 9]
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(self.db.llen('newlist'), len(new_list))
        self.assertEqual(self.db.llen('list'), -1) 

    def test_lappend(self):
        new_list = [4, 5, 6, 7, 8, 9]
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(self.db.lappend('newlist', 0, 3), True)
        self.assertEqual(self.db.lindex('newlist', 0), 7)
        self.assertEqual(self.db.lappend('list', 0, 3), False)
        
    def test_lexists(self):
        new_list = [4, 5, 6, 7, 8, 9]
        self.assertEqual(self.db.lextend('newlist', new_list), True)
        self.assertEqual(self.db.lexists('newlist', 5), True)
        self.assertEqual(self.db.lexists('newlist', 20), False)
        self.assertEqual(self.db.lexists('list', 0), False)


from distutils.core import setup

with open('README.rst') as f:
    long_description = f.read()
    
setup(
  name = 'elara',         
  packages = ['elara'],   
  version = '0.2.1',      
  license='three-clause BSD',        
  description = 'Elara DB is an easy to use, lightweight NoSQL database written for python that can also be used as a fast in-memory cache. Includes various methods to manipulate data structures in-memory, secure database files and export data.',   
  long_description = long_description,
  author = 'Saurabh Pujari',                   
  author_email = 'saurabhpuj99@gmail.com',      
  url = 'https://github.com/saurabh0719/elara',  
  keywords = [
    'database',
    'key-value',
    'storage',
    'file storage',
    'json storage',
    'json database',
    'key-value database' ,
    'nosql',
    'nosql database' 
    'cache',
    'file cache'
    ],   
  install_requires=[
    'cryptography'
    ],
  classifiers=[
    'Development Status :: 4 - Beta',      
    'Intended Audience :: Developers',      
    'Topic :: Database',
    'License :: OSI Approved :: BSD License',   
    'Programming Language :: Python'
  ],
)